# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""Tests for :mod:`requests_scitokens.utils`
"""

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

import pytest

from requests_scitokens import utils as rsutils


@pytest.mark.parametrize(("url", "kwargs", "aud"), (
    # basic
    ("https://example.com/data", {}, "https://example.com"),
    # default scheme
    ("example.com", {}, "https://example.com"),
    # port
    ("https://example.com:443/data/test", {}, "https://example.com"),
    # non-default scheme
    ("http://example.com:443/data/test", {}, "http://example.com"),
    # keyword scheme
    ("example.com:443/data/test", {"scheme": "xroot"}, "xroot://example.com"),
))
def test_default_audience(url, kwargs, aud):
    """Test :func:`requests_scitokens.utils.default_audience`.
    """
    assert rsutils.default_audience(url, **kwargs) == aud



# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""SciToken plugin for Requests.
"""

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

from .auth import HTTPSciTokenAuth
from .requests import (
    delete,
    get,
    head,
    patch,
    post,
    put,
)
from .sessions import (
    Session,
    SessionMixin,
)

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # development mode
    __version__ = ''



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

from pathlib import Path

from sphinx.ext.apidoc import main as apidoc_main

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

from requests_scitokens import __version__ as requests_scitokens_version


# -- Project information -----------------------------------------------------

project = 'requests-scitokens'
copyright = '2024, Cardiff University'
author = 'Duncan Macleod'

# The full version, including alpha/beta/rc tags
release = requests_scitokens_version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_automodapi.automodapi",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'monokai'

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'obj'

# -- Extensions --------------------------------------------------------------

# Intersphinx directory
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "requests": ("https://requests.readthedocs.io/en/stable/", None),
    "scitokens": ("https://scitokens.readthedocs.io/en/stable/", None),
}

# napoleon configuration
napoleon_use_rtype = False

# Don't inherit in automodapi
numpydoc_show_class_members = False
automodapi_inherited_members = False

# autodoc
#autodoc_mock_imports = ['cryptography', 'requests']
autodoc_default_flags = ['members', 'show-inheritance']
autosummary_generate = True



# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""Request API for requests-scitokens.
"""

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

from functools import wraps

import requests.api

from requests_scitokens.auth import HTTPSciTokenAuth
from requests_scitokens.utils import default_audience


@wraps(requests.request)
def request(method, url, *args, auth=None, session=None, **kwargs):
    """Send a SciToken-aware request.

    Parameters
    ----------
    method : `str`
        The method to use.

    url : `str`,
        The URL to request.

    session : `requests.Session`, optional
        The connection session to use, if not given one will be
        created on-the-fly.

    args, kwargs
        All other keyword arguments are passed directly to
        `requests.Session.request`

    Returns
    -------
    resp : `requests.Response`
        the response object

    See also
    --------
    igwn_auth_utils.requests.Session.request
        for information on how the request is performed
    """
    if auth is None:
        auth = HTTPSciTokenAuth(
            token=kwargs.get("token"),
            audience=kwargs.get("audience", default_audience(url)),
        )

    if not session:  # use module
        session = requests
    return session.request(method, url, *args, auth=auth, **kwargs)


def _request_wrapper_factory(method):
    """Factor function to wrap a :mod:`requests` HTTP method to use
    our request function.
    """
    @wraps(getattr(requests.api, method))
    def _request_wrapper(url, *args, session=None, **kwargs):
        return request(method, url, *args, session=session, **kwargs)

    return _request_wrapper


# request methods
delete = _request_wrapper_factory("delete")
get = _request_wrapper_factory("get")
head = _request_wrapper_factory("head")
patch = _request_wrapper_factory("patch")
post = _request_wrapper_factory("post")
put = _request_wrapper_factory("put")



# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""Utilities for requests_scitokens.
"""

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

from requests import __version__ as requests_version
from requests.utils import prepend_scheme_if_needed
from urllib3.util import parse_url

REQUESTS_PREPEND_SCHEME_BROKEN = requests_version < "2.27.0"


def default_audience(url, scheme="https"):
    """Return the expected ``aud`` claim to authorize a request to ``url``.

    Parameters
    ----------
    url : `str`
        The URL that will be requested.

    Returns
    -------
    audience : `str`
        The audience values (`str`) expected for ``url``.

    Examples
    --------
    >>> default_audience(
    ...     "https://storage.example.com:1095//my/data.dat",
    ... )
    'https://storage.example.com'

    Hostnames given without a URL scheme have the ``scheme`` keyword
    prepended to the audience URI.

    >>> default_audience("storage.example.com")
    'https://storage.example.com'
    """
    if REQUESTS_PREPEND_SCHEME_BROKEN and "://" not in url:
        url = f"//{url}"
    else:
        url = prepend_scheme_if_needed(url, scheme)
    parsed = parse_url(prepend_scheme_if_needed(url, scheme))
    return f"{parsed.scheme}://{parsed.hostname}"


def serialize_token(token):
    """Serialise a :class:`~scitokens.SciToken`.

    If ``token`` was parsed from a serialisation, the same serialisation
    will be returned, otherwise the :meth:`~scitokens.SciToken.serialize`
    method of ``token`` will be called.
    """
    return (
        token._serialized_token
        or token.serialize().decode("utf-8")
    )



# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""Utilities for the `requests-scitokens`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import time

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key

import pytest

from scitokens import SciToken


# -- old pytest fixtures --------------

if pytest.__version__ < "3.9.0":  # RL8
    from pathlib import Path
    from tempfile import TemporaryDirectory

    @pytest.fixture
    def tmp_path():
        with TemporaryDirectory() as tmpdir:
            yield Path(tmpdir)


# -- SciTokens fixtures ---------------

ISSUER = "local"
_SCOPE_PATH = "/igwn_auth_utils"
READ_AUDIENCE = "igwn_auth_utils"
READ_SCOPE = "read:{}".format(_SCOPE_PATH)
WRITE_AUDIENCE = "igwn_auth_utils2"
WRITE_SCOPE = "write:{}".format(_SCOPE_PATH)


@pytest.fixture(scope="session")  # one per suite is fine
def private_key():
    """Generate a private RSA key with which to sign a token.
    """
    return generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )


@pytest.fixture(scope="session")
def public_key(private_key):
    """The public key for the private_key.
    """
    return private_key.public_key()


@pytest.fixture(scope="session")
def public_pem(public_key):
    """The public_key in PEM format.
    """
    return public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )


def _create_token(
    key=None,
    iss=ISSUER,
    aud=READ_AUDIENCE,
    scope=READ_SCOPE,
    **kwargs
):
    """Create a token.
    """
    if key:
        from scitokens.utils.keycache import KeyCache
        keycache = KeyCache.getinstance()
        keycache.addkeyinfo(iss, "test_key", key.public_key())
    now = int(time.time())
    token = SciToken(key=key, key_id="test_key")
    token.update_claims({
        "iat": now,
        "nbf": now,
        "exp": now + 86400,
        "iss": iss,
        "aud": aud,
        "scope": scope,
    })
    token.update_claims(kwargs)
    return token


def _write_token(token, path):
    """Write a token to a file.
    """
    with open(path, "wb") as file:
        file.write(token.serialize(lifetime=86400))


@pytest.fixture
def rtoken(private_key):
    """A token with the ``READ_SCOPE``.
    """
    return _create_token(
        key=private_key,
        scope=READ_SCOPE,
    )


@pytest.fixture
def wtoken(private_key):
    """A token with the ``WRITE_SCOPE``.
    """
    return _create_token(
        key=private_key,
        aud=WRITE_AUDIENCE,
        scope=WRITE_SCOPE,
    )


@pytest.fixture
def rtoken_path(rtoken, tmp_path):
    """The path of a file containing the serialised ``READ_SCOPE`` token.
    """
    token_path = tmp_path / "token.use"
    _write_token(rtoken, token_path)
    return token_path



# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""Session API for requests-scitokens.
"""

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

import requests

from requests_scitokens.auth import HTTPSciTokenAuth


class SessionMixin():
    def __init__(self, *args, auth=None, token=None, **kwargs):
        if auth is None:
            auth = HTTPSciTokenAuth(token=token)
        super().__init__(*args, auth=auth, **kwargs)


class Session(SessionMixin, requests.Session):
    pass



# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""Tests for :mod:`requests_scitokens.auth`
"""

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

import os
from unittest import mock

import requests

from scitokens import SciToken

from requests_scitokens import auth
from requests_scitokens.utils import serialize_token

# -- mock helpers ---------------------

deserialize = SciToken.deserialize


def _insecure_deserialize_factory(key):
    """Generate an insecure version of SciToken.deserialize.

    Hardcodes ``insecure=True, public_key={key}``.
    """
    def _insecure_deserialize(
        tokenstr,
        audience,
        require_key,
        insecure,
        public_key,
    ):
        return deserialize(
            tokenstr,
            audience,
            require_key,
            insecure=True,
            public_key=key,
        )
    return _insecure_deserialize


def _token_response(request, context):
    """Text response based on whether a token was received or not.
    """
    if request.headers.get("Authorization", "").startswith("Bearer"):
        context.status_code = 200
        return "Success"
    context.status_code = 401
    return "Unauthorized"


def assert_tokens_equal(a, b):
    assert dict(a.claims()) == dict(b.claims())


class MockRequest(mock.MagicMock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = {}


class TestHTTPSciTokenAuth:
    Auth = auth.HTTPSciTokenAuth

    def test_init(self):
        auth = self.Auth()
        assert auth.token is None
        assert auth.audience is None

    def test_eq(self):
        a = self.Auth(token=None, audience="ANY")
        b = self.Auth(token=None, audience="ANY")
        assert a == b

    def test_neq(self):
        a = self.Auth(token=None, audience="ANY")
        b = self.Auth(token=None, audience="https://example.com")
        assert a != b

    @mock.patch("requests_scitokens.auth.SciToken.discover")
    def test_token_header_empty(self, find_token):  # noqa: F811
        """Test that the auth class handles no tokens properly.
        """
        find_token.return_value = None
        req = MockRequest()
        auth = self.Auth()
        assert auth(req).headers.get("Authorization") is None

    def test_token_header(self, rtoken):  # noqa: F811
        """Test that the auth class finds the token and serialises
        it into a header properly.
        """
        auth = self.Auth(token=rtoken)
        req = MockRequest()
        assert auth(req).headers["Authorization"] == (
            f"Bearer {serialize_token(rtoken)}"
        )

    @mock.patch.dict(os.environ)
    def test_find_token(self, rtoken, rtoken_path, public_pem):
        """Test the the `find_token` method works.
        """
        os.environ["BEARER_TOKEN_FILE"] = str(rtoken_path)
        auth = self.Auth()
        token = auth.find_token(insecure=True, public_key=public_pem)
        assert_tokens_equal(token, rtoken)

    def test_handle_401_no_challenge(self, requests_mock):
        """Test that a 401 without a challenge is handled properly.
        """
        # mock a request that doesn't respond with a Bearer challenge
        # in the WWW-Authenticate header
        requests_mock.get(
            "https://example.com/",
            status_code=401,
            text="Access denied",
        )
        # and check that we propagate the response
        resp = requests.get(
            "https://example.com",
            auth=self.Auth(),
        )
        assert resp.status_code == 401
        assert resp.text == "Access denied"

    @mock.patch.dict(os.environ)
    def test_handle_401_challenge(self, rtoken, public_pem, requests_mock):
        """Test that a 401 without a challenge is handled properly.
        """
        os.environ["BEARER_TOKEN"] = serialize_token(rtoken)
        requests_mock.get(
            "https://example.com/",
            [
                # initial response issues bearer challenge
                {
                    "status_code": 401,
                    "headers": {"WWW-Authenticate": "Bearer"},
                },
                # second response should pass if token was received
                {"text": _token_response},
            ],
        )
        with mock.patch(
            "requests_scitokens.auth.SciToken.deserialize",
            _insecure_deserialize_factory(public_pem),
        ):
            assert requests.get(
                "https://example.com",
                auth=self.Auth(),
            ).status_code == 200



# -*- python -*-
# Copyright (C) 2024 Cardiff University
# SPDX-License-Identifier: Apache-2.0

"""Auth plugin for SciToken requests.
"""

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

from requests.auth import AuthBase as _AuthBase
from requests.utils import (
    parse_dict_header,
    parse_list_header,
)

from scitokens import SciToken

from requests_scitokens.utils import (
    default_audience,
    serialize_token,
)


def _www_authenticate_bearer_params(response):
    """Return the WWW-Authenticate Bearer response auth-params if found.

    See RFC 6750, section 3.

    Returns
    -------
    params : `False`, `dict`
        Either `False` if the ``www-authenticate`` header is not present,
        or doesn't challenege Bearer auth, or a `dict` (possible empty)
        of params to use in the Bearer challenge.
    """
    try:
        wwwauth = response.headers["www-authenticate"]
    except KeyError:
        return False

    for authopt in parse_list_header(wwwauth):
        # if header is just 'Bearer', then return empty dict
        if authopt == "Bearer":
            return {}
        # otherwise if header starts with Bearer and some whitespace
        # extract the params  -- this just avoids falsely matching
        # something like 'Bearer2', which we don't understand.
        if authopt.startswith("Bearer "):  # with params
            return parse_dict_header(authopt[6:])

    # no match
    return False


class HTTPSciTokenAuth(_AuthBase):
    """Auth handler for SciTokens.

    Parameters
    ----------
    token : `scitokens.SciToken`, `None`
        The `~scitokens.SciToken` to use for bearer authorisation.
        Pass `None` to dynamically discover a token for each request.

    audience : `str`
        The ``aud`` claim to require when discovering tokens.

    force : `bool`
        If `True`, preemptively discover a token and generate an
        ``Authorization`` header for all requests.

        If `False`, wait for the remote server to respond with an
        authorisation challenge via a 401 ``Unauthorized`` response
        that includes a ``WWW-Authenticate`` header.
    """
    def __init__(
        self,
        token=None,
        audience=None,
        force=False,
    ):
        self.token = token
        self.audience = audience
        self.force = force

    def __eq__(self, other):
        return all([
            self.token == getattr(other, "token", None),
            self.audience == getattr(other, "audience", None),
            self.force == getattr(other, "force", False),
        ])

    def __ne__(self, other):
        return not self == other

    @staticmethod
    def _auth_header_str(token, auth_scheme="Bearer"):
        """Serialise a `scitokens.SciToken` and format an Authorization header.

        Parameters
        ----------
        token : `scitokens.SciToken`, `str`, `bytes`
            The token to serialize, or an already serialized representation.

        auth_scheme : `str`
            The authorisation scheme to use, defaults to ``"Bearer"``.
        """
        if not isinstance(token, (str, bytes)):
            token = serialize_token(token)
        return f"{auth_scheme} {token}"

    def find_token(
        self,
        url=None,
        error=True,
        find_func=SciToken.discover,
        **discover_kwargs,
    ):
        """Find a bearer token for authorization.

        Parameters
        ----------
        url : `str`
            The URL that will be queried.

        error : `bool`
            If `True`, `raise` exceptions, otherwise return `None`.

        find_func : `callable`
            The function to call to discover SciTokens, defaults
            to :meth:`~scitokens.SciToken.discover`.

        discover_kwargs
            Other keyword arguments to pass to ``find_func``.

        Returns
        -------
        token : `scitokens.SciToken`, `None`
            The discovered token, or `None` if a token is not foudn
            and ``error=False`` is given.

        Raises
        ------
        OSError
            If ``error=True`` is given (default) and a token discovery
            fails.
        """
        audience = self.audience
        if audience is None and url is not None:
            audience = default_audience(url)
        try:
            return find_func(
                audience=audience,
                **discover_kwargs,
            )
        except OSError:  # failed to discover token
            if error:
                raise
            return None

    def generate_bearer_header(self, response=None):
        """Generate a bearer token header, possibly based on a response.
        """
        token = self.token
        error = response is not None or token is True
        if (
            response is not None
            or token in (None, True)
        ):
            # try and find a token
            token = self.find_token(
                url=getattr(response, "url", None),  # allow r as Session
                error=error,
            )

        # if we ended up with a token, generate the header content
        if token:
            return self._auth_header_str(token)

    def authenticate_bearer(self, response, **kwargs):
        """Re-send a request in response to a Bearer challenge.
        """
        try:
            auth_header = self.generate_bearer_header(response)
        except ValueError:
            # return original response
            return response

        # consume the content so that we can reuse the connection
        response.content
        response.raw.release_conn()

        # retry the same request, using the same connection
        request = response.request.copy()
        request.headers["Authorization"] = auth_header
        new = response.connection.send(request, **kwargs)
        new.history.append(response)
        return new

    def handle_401(self, response, **kwargs):
        """Handle 401 response.
        """
        params = _www_authenticate_bearer_params(response)
        if isinstance(params, dict):
            kwargs.update(params)
            return self.authenticate_bearer(response, **kwargs)
        return response

    def handle_response(self, response, **kwargs):
        """Attempt to handle a response that asks for Bearer auth.
        """
        num_401s = kwargs.pop("num_401s", 0)
        if response.status_code == 401 and num_401s < 2:
            new = self.handle_401(response, **kwargs)
            num_401s += 1
            return self.handle_response(new, num_401s=num_401s, **kwargs)

        return response

    def __call__(self, request):
        """Augment the `Request` ``request`` with an ``Authorization`` header.
        """
        token = self.token
        if self.force:
            # try and find a token
            token = self.find_token(
                url=request.url,
                error=True,
            )

        # if we ended up with a header, store it in the request.
        if token:
            request.headers["Authorization"] = self._auth_header_str(token)

        # register our reponse handler
        request.register_hook("response", self.handle_response)

        return request
